from django.test import TestCase
from djeddit.models import Topic, Thread, Post

# Create your tests here.


class TopicModelTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.topic = Topic.objects.create(title='Test Topic')
        # create 3 threads
        cls.threadCount = 3
        cls.threads = [Thread.objects.create(title='Thread %s' % i,
                                             topic=cls.topic,
                                             op=Post.objects.create()) for i in range(1, cls.threadCount + 1)]

    def testThreadCount(self):
        self.assertEqual(self.topic.getThreadCount(), self.threadCount)


class ThreadModelTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.thread = Thread.objects.create(title='Test Thread',
                                           topic=Topic.objects.create(title='Test Topic'),
                                           op=Post.objects.create())

    def testThreadDelete(self):
        uid = self.thread.op.uid
        self.thread.delete()
        self.assertRaises(Thread.DoesNotExist, Thread.objects.get, title='Test Thread')
        self.assertRaises(Post.DoesNotExist, Post.objects.get, uid=uid)


class PostModelTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.post = Post.objects.create()
        cls.thread = Thread.objects.create(title='Test Thread', topic=Topic.objects.create(title='Test Topic'), op=cls.post)
        cls.replies = [Post.objects.create(parent=cls.post) for _ in range(1, 4)]

    def testGetReplies(self):
        self.assertListEqual(self.replies, list(self.post.getReplies()))

    def getThread(self):
        self.assertEqual(self.thread, self.post.getThread())



from django.test import TestCase
from django.urls import reverse
from djeddit.utils.base_tests import TestCalls, createUser

from djeddit.models import Topic, Thread, Post, gen_uuid, UserPostVote


# Create your tests here.


class CreateThreadTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/create_thread.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test_Topic')
        cls.url = reverse('createThread', args=[cls.topic.title])

    def testLoads(self):
        self._test_call_view_loads(self.url, {})

    def _testSubmit(self, data):
        self._test_call_view_submit(self.url, data, redirect=True)
        thread = Thread.objects.get(topic=self.topic, title=data['thread-title'])
        self.assertEqual(thread.op.content, data['post-content'])

    def testSubmitUnauthenticated(self):
        data = {'post-content': 'content', 'thread-title': 'Thread1', 'thread-url': ''}
        self._testSubmit(data)

    def testSubmitAuthenticated(self):
        data = {'post-content': 'content', 'thread-title': 'Thread1', 'thread-url': ''}
        self.login()
        self._testSubmit(data)

class TopicsPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/topics.html')

    def testLoads(self):
        url = reverse('topics')
        self._test_call_view_loads(url)


class TopicPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/topic.html')

    @classmethod
    def setUpTestData(cls):
        cls.topic = Topic.objects.create(title='Test_Topic')

    def testLoads(self):
        url = reverse('topicPage', args=[self.topic.title])
        self._test_call_view_loads(url)

    def testRedirects(self):
        url = reverse('topicPage', args=['Fake_Topic'])
        redirected_url = reverse('topics')
        self._test_call_view_redirects(url, redirected_url)


class ThreadPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/thread.html')

    @classmethod
    def setUpTestData(cls):
        cls.topic = Topic.objects.create(title='Test_Topic')
        cls.thread = Thread.objects.create(title='Test_Thread', topic=cls.topic, op=Post.objects.create())

    def testLoads(self):
        url = reverse('threadPage', args=[self.topic.title, self.thread.id])
        self._test_call_view_loads(url)

    def testWrongTopic(self):
        url = reverse('threadPage', args=['Fake_Topic', self.thread.id])
        redirected_url = reverse('topics')
        self._test_call_view_redirects(url, redirected_url)

    def testWrongThread(self):
        url = reverse('threadPage', args=[self.topic.title, self.thread.id + 1])
        redirected_url = reverse('topics')
        self._test_call_view_redirects(url, redirected_url)


class ReplyPostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/reply_form.html')

    @classmethod
    def setUpTestData(cls):
        cls.post = Post.objects.create()
        topic = Topic.objects.create(title='Test_Topic')
        Thread.objects.create(title='Test_Thread', topic=topic, op=cls.post)
        cls.url = reverse('replyPost', args=[cls.post.uid])

    def testLoads(self):
        self._test_call_view_loads(self.url)

    def testSubmit(self):
        self._test_call_view_submit(self.url, dict(content='replied'), redirect=True)
        Post.objects.get(parent=self.post, content='replied') # assert object exists

    def testUnknownUid(self):
        uid = gen_uuid()
        url = reverse('replyPost', args=[str(uid)])
        self._test_call_view_code(url, 404)

    def testNoThread(self):
        post = Post.objects.create()
        url = reverse('replyPost', args=[str(post.uid)])
        self._test_call_view_code(url, 404)


class EditPostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/edit_post.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.post = Post.objects.create(created_by=cls.user)
        topic = Topic.objects.create(title='Test_Topic')
        Thread.objects.create(title='Test_Thread', topic=topic, op=cls.post)
        cls.url = reverse('editPost', args=[cls.post.uid])

    def testLoads(self):
        self.login()
        self._test_call_view_loads(self.url)

    def testSubmit(self):
        self.login()
        self._test_call_view_submit(self.url, dict(content='edited'), redirect=True)
        self.post.refresh_from_db()
        self.assertEqual(self.post.content, 'edited')

    def testUnknownUid(self):
        uid = gen_uuid()
        self.login()
        url = reverse('editPost', args=[str(uid)])
        self._test_call_view_code(url, 404)

    def testNoThread(self):
        post = Post.objects.create()
        url = reverse('editPost', args=[str(post.uid)])
        self.login()
        self._test_call_view_code(url, 404)

    def testEditAnotherPost(self):
        user = createUser('user1', 'user1@example.com', 'pass')
        post = Post.objects.create(created_by=user)
        url = reverse('editPost', args=[str(post.uid)])
        self.login()
        self._test_call_view_code(url, 404)


class VotePostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.post = Post.objects.create()
        cls.url = reverse('votePost')

    def _testVote(self, submit_val, expected_score, expected_votePost_val):
        self._test_call_view_submit(self.url, dict(post=self.post.uid, vote=submit_val))
        user_vote = UserPostVote.objects.get(user=VotePostTest.user, post=self.post)  # assert object exists
        self.assertEqual(user_vote.val, expected_votePost_val)
        self.post.refresh_from_db()
        self.assertEqual(self.post.score, expected_score)

    def testSubmit(self):
        self.login()
        self.assertEqual(self.post.score, 0)
        self._testVote(1, 1, 1) # test upvote
        self._testVote(2, 1, 1) # test vote value above 1
        self._testVote(-1, -1, -1) # test downvote
        self._testVote(-2, -1, -1) # test downvote value below -1

    def testRequiresLogin(self):
        self._test_call_view_require_login(self.url, dict(post=self.post.uid, vote=1))

    def testVoteOwnPost(self):
        self.login()
        post = Post.objects.create(created_by=VotePostTest.user)
        response = self.client.post(self.url, dict(post=post.uid, vote=1))
        self.assertEqual(response.status_code, 403)


class UserSummaryTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_summary.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test_Topic')
        cls.url = reverse('userSummary', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        op = Post.objects.create(created_by=self.user)
        Thread.objects.create(topic=self.topic, title='Test Thread', op=op)
        Post.objects.create(created_by=self.user)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)


class UserThreadsPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_threads.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test_Topic')
        cls.url = reverse('userThreads', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        op = Post.objects.create(created_by=self.user, content='op content')
        Thread.objects.create(topic=self.topic, title='Test Thread', op=op)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)


class UserRepliesPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_threads.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        topic = Topic.objects.create(title='Test_Topic')
        cls.thread = Thread.objects.create(topic=topic, title='Test Thread', op=Post.objects.create())
        cls.url = reverse('userThreads', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        Post.objects.create(parent=self.thread.op)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)



# -*- coding: utf-8 -*-
from django.conf.urls import url
from . import views

urlpatterns = [
    url(r'^$', views.homePage, name='home'),
    url(r'^topics/?$', views.topicsPage, name='topics'),
    url(r'^topics/(\w+)?/?$', views.topicPage, name='topicPage'),
    url(r'^topics/(\w+)/newthread/?$', views.createThread, name='createThread'),
    url(r'^topics/(\w+)/(\d+)?/?$', views.threadPage, name='threadPage'),
    url(r'^reply_post/([\w\-]{36})?/?$', views.replyPost, name='replyPost'),
    url(r'^edit_post/([\w\-]{36})?/?$', views.editPost, name='editPost'),
    url(r'^vote_post/?$', views.votePost, name='votePost'),
    url(r'^user/(.+)/summary/?$', views.userSummary, name='userSummary'),
    url(r'^user/(.+)/threads/?$', views.userThreadsPage, name='userThreads'),
    url(r'^user/(.+)/replies/?$', views.userRepliesPage, name='userReplies'),
]



from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.shortcuts import render, redirect
from django.http import Http404, HttpResponseForbidden

from djeddit.forms import TopicForm, ThreadForm, PostForm
from djeddit.models import Topic, Thread, Post, UserPostVote
from djeddit.templatetags.djeddit_tags import postScore
from django.contrib.auth.models import User


# Create your views here.

# @user_passes_test(lambda user: user.is_superuser)

def createThread(request, topic_title=None):
    if topic_title:
        try:
            if request.method == 'POST':
                topic = Topic.objects.get(title=topic_title)
                threadForm = ThreadForm(request.POST, prefix='thread')
                postForm = PostForm(request.POST, prefix='post')
                if threadForm.is_valid() and postForm.is_valid():
                    thread = threadForm.save(commit=False)
                    post = postForm.save(commit=False)
                    thread.op = post
                    thread.topic = topic
                    thread.save()
                    post.thread = thread
                    if request.user.is_authenticated:
                        post.created_by = request.user
                    post.save()
                    return redirect('threadPage', topic.title, thread.id)
            else:
                threadForm = ThreadForm(prefix='thread')
                postForm = PostForm(prefix='post')
            context = dict(threadForm=threadForm, postForm=postForm)
            return render(request, 'djeddit/create_thread.html', context)
        except Topic.DoesNotExist:
            pass
    return redirect('topics')


def topicsPage(request):
    topics = Topic.objects.all()
    showForm = False
    if request.method == 'POST':
        topicForm = TopicForm(request.POST)
        if topicForm.is_valid():
            topicForm.save()
            return redirect('topics')
        showForm = True
    else:
        topicForm = TopicForm()
    context = dict(topics=topics, topicForm=topicForm, showForm=showForm)
    return render(request, 'djeddit/topics.html', context)


def topicPage(request, topic_title=None):
    if topic_title:
        try:
            topic = Topic.objects.get(title=topic_title)
            threads = Thread.objects.filter(topic=topic)
            context = dict(topic=topic, threads=threads, showCreatedBy=True, showTopic=False)
            return render(request, 'djeddit/topic.html', context)
        except Topic.DoesNotExist:
            pass
    return redirect('topics')


def threadPage(request, topic_title='', thread_id=''):
    if topic_title and thread_id:
        try:
            thread = Thread.objects.get(id=thread_id)
            if thread.topic.title == topic_title:
                thread.views += 1
                thread.save()
                context = dict(thread=thread, nodes=thread.op.getReplies())
                return render(request, 'djeddit/thread.html', context)
        except Thread.DoesNotExist:
            pass
    return redirect('topics')


def homePage(request):
    # context = sorted(dict(request.GET, **request.POST).items())
    context = dict(request.GET, **request.POST)
    return render(request, 'djeddit/home.html', context)


def replyPost(request, post_uid=''):
    try:
        repliedPost = Post.objects.get(uid=post_uid)
        thread = repliedPost.getThread()
    except (Post.DoesNotExist, Thread.DoesNotExist):
        raise Http404
    repliedUser = repliedPost.created_by.username if repliedPost.created_by else 'guest'
    if request.method == 'POST':
        postForm = PostForm(request.POST)
        if postForm.is_valid():
            post = postForm.save(commit=False)
            post.parent = repliedPost
            if request.user.is_authenticated:
                post.created_by = request.user
            post.save()
            repliedPost.children.add(post)
        return redirect('threadPage', thread.topic.title, thread.id)
    else:
        postForm = PostForm()
        postForm.fields['content'].label = ''
        context = dict(postForm=postForm, thread_id=thread.id, post_uid=post_uid, repliedUser=repliedUser)
        return render(request, 'djeddit/reply_form.html', context)


def editPost(request, post_uid=''):
    try:
        post = Post.objects.get(uid=post_uid)
        thread = post.getThread()
    except (Post.DoesNotExist, Thread.DoesNotExist):
        raise Http404
    if request.user != post.created_by and not request.user.is_superuser:
        return HttpResponseForbidden()
    if request.method == 'POST':
        postForm = PostForm(request.POST, instance=post)
        if postForm.is_valid():
            postForm.save()
        return redirect('threadPage', thread.topic.title, thread.id)
    else:
        postForm = PostForm(vars(post))
        postForm.fields['content'].label = ''
        context = dict(postForm=postForm, post_uid=post.uid)
        return render(request, 'djeddit/edit_post.html', context)


@login_required
def votePost(request):
    post_uid = request.POST['post']
    vote_val = request.POST['vote']
    post = Post.objects.get(uid=post_uid)
    if post.created_by != request.user or request.user.is_superuser:
        try:
            userPostVote = UserPostVote.objects.get(user=request.user, post=post)
            oldval = userPostVote.val
            userPostVote.val = max(min(int(vote_val), 1), -1)
            userPostVote.save()
            voteDelta = userPostVote.val - oldval
        except UserPostVote.DoesNotExist:
            userPostVote = UserPostVote.objects.create(user=request.user, post=post, val=max(min(int(vote_val), 1), -1))
            voteDelta = userPostVote.val
        post.score += voteDelta
        post.save()
        scoreStr = postScore(post.score)
        return JsonResponse(dict(scoreStr=scoreStr, score=post.score))
    else:
        return HttpResponseForbidden()


def userSummary(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    threads = Thread.objects.filter(op__created_by=user)
    for t in threads:
        t.modified_on = t.op.modified_on
    replies = Post.objects.filter(created_by=user).exclude(uid__in=(t.op.uid for t in threads))
    context = dict(items=sorted(list(threads) + list(replies), key=lambda n: n.modified_on, reverse=True),
                   tCount=threads.count(),
                   rCount=replies.count(),
                   tPoints=(sum(t.op.score for t in threads)),
                   rPoints=(sum(r.score for r in replies)),
                   pageUser=user)
    return render(request, 'djeddit/user_summary.html', context)


def userThreadsPage(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    created_threads = Thread.objects.filter(op__created_by=user)
    context = dict(threads=created_threads, showCreatedBy=False, showTopic=True, pageUser=user)
    return render(request, 'djeddit/user_threads.html', context)


def userRepliesPage(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    replies = Post.objects.filter(created_by=user, parent__isnull=False)
    context = dict(replies=replies, pageUser=user)
    return render(request, 'djeddit/user_replies.html', context)



import uuid
from django.conf import settings
from django.db import models
from mptt.models import MPTTModel, TreeForeignKey


# Create your models here.

def gen_uuid():
    return uuid.uuid4()


class IntegerRangeField(models.IntegerField):
    def __init__(self, verbose_name=None, name=None, min_value=None, max_value=None, **kwargs):
        self.min_value, self.max_value = min_value, max_value
        models.IntegerField.__init__(self, verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {'min_value': self.min_value, 'max_value': self.max_value}
        defaults.update(kwargs)
        return super(IntegerRangeField, self).formfield(**defaults)


class NamedModel(models.Model):
    class Meta:
        abstract = True

    def getName(self):
        return self.__class__.__name__


class Topic(NamedModel):
    title = models.CharField(max_length=120, blank=False, unique=True)

    def getThreadCount(self):
        return Thread.objects.filter(topic=self).count()


class Thread(NamedModel):
    title = models.CharField(max_length=120, blank=False)
    url = models.URLField(max_length=120, blank=True, default='')
    views = models.IntegerField(blank=True, default=0)
    topic = models.ForeignKey(Topic, on_delete=models.CASCADE)
    op = models.ForeignKey('Post', related_name='+', on_delete=models.CASCADE)

    def delete(self, *args, **kwargs):
        try:
            self.op.delete()
        except Post.DoesNotExist:
            pass
        super(Thread, self).delete(*args, **kwargs)


class Post(MPTTModel, NamedModel):
    content = models.TextField(blank=True, default='')
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE, related_name='+')
    created_on = models.DateTimeField(auto_now_add=True, auto_now=False)
    modified_on = models.DateTimeField(auto_now_add=False, auto_now=True)
    parent = TreeForeignKey('self', null=True, blank=True, related_name='children', db_index=True, on_delete=models.CASCADE)
    score = models.IntegerField(blank=True, default=0)
    uid = models.UUIDField(max_length=8, primary_key=True, default=gen_uuid, editable=False)

    class MPTTMetta:
        order_insertion_by = ['created_on']

    def getReplies(self):
        replies = Post.objects.filter(parent=self.uid)
        for reply in replies:
            replies |= reply.getReplies()
        return replies

    def getThread(self):
        post = self
        while post.parent:
            post = post.parent
        return Thread.objects.get(op=post)


class UserPostVote(NamedModel):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='+')
    post = models.ForeignKey(Post, related_name='+')
    val = IntegerRangeField(blank=True, default=0, min_value=-1, max_value=1)



from django.contrib.auth.models import User


def createUser(username, email, password):
    user = User(username=username, email=email)
    user.set_password(password)
    user.save()
    return user

class TestCalls(object):

    def __init__(self, template=''):
        self.template = template
        self.user = None

    def login(self):
        if self.username and self.password:
            self.client.login(username=self.username, password=self.password)

    @classmethod
    def _setup_user(cls, username, email, password):
        cls.user = createUser(username, email, password)
        cls.username = username
        cls.password = password

    def _test_call_view_loads(self, url, data=None):
        data = data or {}
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(self.template)

    def _test_call_view_submit(self, url, data=None, redirect=False):
        data = data or {}
        response = self.client.post(url, data)
        expected_code = 302 if redirect else 200
        self.assertEqual(response.status_code, expected_code)
        if self.template:
            self.assertTemplateUsed(self.template)

    def _test_call_view_require_login(self, url, data=None):
        data = data or {}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

    def _test_call_view_redirects(self, url, redirected_url, data=None):
        data = data or {}
        response = self.client.get(url, data)
        self.assertRedirects(response, redirected_url)

    def _test_call_view_code(self, url, code, data=None):
        data = data or {}
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, code)
